from dataclasses import dataclass
import math

from .primitives import Angle
from .zodiac import Zodiac, ZodiacConstell

@dataclass
class EclCoord:
    """Ecliptic coordinate"""

    latitude: Angle
    longitude: Angle

    def __init__(self, longitude: float, latitude: float):
        self.longitude = Angle(longitude)
        self.latitude = Angle(latitude)

    def __xor__(self, other) -> Angle:
        if self == other:
            return Angle(0)
        long1 = self.longitude.radians()
        long2 = other.longitude.radians()
        lat1 = self.latitude.radians()
        lat2 = other.latitude.radians()
        dalpha = long1 - long2
        cosd = math.sin(lat1) * math.sin(lat2) + math.cos(lat1) * math.cos(lat2) * math.cos(dalpha)
        ang = math.acos(cosd)
        if ang > math.pi:
            ang = 2 * math.pi - ang
        return Angle(radians=ang)

    def sign_pos(self) -> (Zodiac, float):
        sign = Zodiac.from_longitude(self.longitude)
        return sign, self.longitude.degrees - sign.offset

    def constell_pos(self) -> (ZodiacConstell, float):
        constell = ZodiacConstell.from_longitude(self.longitude)
        pos = self.longitude.degrees
        if constell.upto_lng < constell.from_lng < pos:
            offset = pos - constell.from_lng
        elif pos < constell.upto_lng < constell.from_lng:
            offset = pos
        else:
            offset = pos - constell.from_lng
        return constell, offset


@dataclass
class EquatorCoord:
    """Equatorial coordinate"""

    ra: Angle
    decl: Angle

    def __init__(self, ra: float, decl: float):
        self.ra = Angle(ra)
        self.decl = Angle(decl)


@dataclass
class HorCoord:
    """Horizontal coordinate"""

    azimuth: Angle
    altitude: Angle

    def __init__(self, azimuth: float, altitude: float):
        self.azimuth = Angle(azimuth)
        self.altitude = Angle(altitude)

    def house_pos(self) -> (int, float):
        alt = self.altitude.radians()
        azimuth = self.azimuth.radians()
        angle = math.atan2(math.tan(alt), math.cos(azimuth)) / math.pi * 180.0 - 90.0
        if angle < 0:
            angle += 360
        house13 = int(math.floor(angle * 13.0 / 360.0)) + 1
        house_pos = angle * 13.0 / 360.0 + 1.0 - house13
        house_pos *= 180.0 / math.pi
        return house13, house_pos



from abc import ABC, abstractmethod
from datetime import datetime, time, timedelta
import swisseph as swe

from . import GeoLocation, EclCoord, EquatorCoord, HorCoord


class Celestial(ABC):
    """Abstract class for celestial objects whose location can be computed"""

    def ecl_coord(self, time: datetime, location: GeoLocation) -> EclCoord:
        swe.set_topo(location.longitude.degrees, location.latitude.degrees)
        jd = swe.julday(time.year, time.month, time.day, time.hour + time.minute / 60.)
        return self.swe_ecl_coord(jd)

    def equator_coord(self, time: datetime, location: GeoLocation) -> EquatorCoord:
        swe.set_topo(location.longitude.degrees, location.latitude.degrees)
        jd = swe.julday(time.year, time.month, time.day, time.hour + time.minute / 60.)
        return self.swe_equator_coord(jd)

    def hor_coord(self, time: datetime, location: GeoLocation) -> HorCoord:
        swe.set_topo(location.longitude.degrees, location.latitude.degrees)
        jd = swe.julday(time.year, time.month, time.day, time.hour + time.minute / 60.)
        coord = self.swe_equator_coord(jd)
        geopos = (location.longitude.degrees, location.latitude.degrees, 0.0)
        pos = (coord.ra.degrees, coord.decl.degrees, 0.0)
        atpress = 0
        attemp = 0
        (azimuth, true_alt, app_alt) = swe.azalt(jd, swe.EQU2HOR, geopos, atpress, attemp, pos)
        return HorCoord(azimuth, true_alt)

    def transits(self, time: datetime, location: GeoLocation):
        return {
            'rise': self.rises(time, location),
            'set': self.sets(time, location),
            'mc': self.mc_trans(time, location),
            'ic': self.ic_trans(time, location),
        }

    def rises(self, time: datetime, location: GeoLocation):
        return self.__rise_trans(time, location, swe.CALC_RISE)

    def sets(self, time: datetime, location: GeoLocation):
        return self.__rise_trans(time, location, swe.CALC_SET)

    def mc_trans(self, time: datetime, location: GeoLocation):
        return self.__rise_trans(time, location, swe.CALC_MTRANSIT)

    def ic_trans(self, time: datetime, location: GeoLocation):
        return self.__rise_trans(time, location, swe.CALC_ITRANSIT)

    def __rise_trans(self, when: datetime, location: GeoLocation, rsmi: int):
        if self.is_focal_point():
            return None
        tjdut = swe.julday(when.year, when.month, when.day, 0.)
        flags = swe.FLG_SWIEPH | swe.FLG_TOPOCTR
        rsmi |= swe.BIT_DISC_CENTER | swe.BIT_FIXED_DISC_SIZE | swe.BIT_NO_REFRACTION | swe.BIT_ASTRO_TWILIGHT
        lon = location.longitude.degrees
        lat = location.latitude.degrees
        alt = 0.0
        atpress = 0
        attemp = 0
        (found, (jultime, _, _, _, _, _, _, _, _, _)) = swe.rise_trans(
            tjdut, self.swe_id(), rsmi, (lon, lat, alt), atpress, attemp, flags
        )
        if found != 0:
            return None
        (year, month, day, tm) = swe.revjul(jultime)
        if year != when.year or month != when.month or day != when.day:
            return None
        hours = int(tm)
        minutes = (tm - hours) * 60.
        seconds = (tm - hours - int(minutes) / 60) * 60 * 60
        return timedelta(hours=hours, minutes=int(minutes), seconds=int(seconds))

    @abstractmethod
    def swe_id(self):
        pass

    @abstractmethod
    def is_fixed(self) -> bool:
        pass

    @abstractmethod
    def is_focal_point(self) -> bool:
        pass

    @abstractmethod
    def swe_ecl_coord(self, jd) -> EclCoord:
        pass

    @abstractmethod
    def swe_equator_coord(self, jd) -> EquatorCoord:
        pass


class Planet(Celestial):
    """Planets (moving physical bodies)"""

    def __init__(self, name: str, swe_code: int):
        self.name = name
        self.__swe_code = swe_code

    def swe_id(self):
        return self.__swe_code

    def is_fixed(self) -> bool:
        return False

    def is_focal_point(self) -> bool:
        return False

    def swe_ecl_coord(self, jd) -> EclCoord:
        (ecl, _) = swe.calc_ut(jd, self.__swe_code, swe.FLG_SWIEPH | swe.FLG_TOPOCTR)
        return EclCoord(ecl[0], ecl[1])

    def swe_equator_coord(self, jd) -> EquatorCoord:
        (equator, _) = swe.calc_ut(jd, self.__swe_code, swe.FLG_SWIEPH | swe.FLG_TOPOCTR | swe.FLG_EQUATORIAL)
        return EquatorCoord(equator[0], equator[1])


class BlackSun(Celestial):
    """Second focal point of some planet orbit"""

    def __init__(self, name: str, swe_code: int):
        self.name = name
        self.__swe_code = swe_code

    def swe_id(self) -> int:
        return self.__swe_code

    def is_fixed(self) -> bool:
        return False

    def is_focal_point(self) -> bool:
        return True

    def swe_ecl_coord(self, jd) -> EclCoord:
        (_, _, _, ecl) = swe.nod_aps_ut(jd, self.__swe_code,
                                        swe.NODBIT_FOPOINT | swe.NODBIT_OSCU,
                                        swe.FLG_SWIEPH | swe.FLG_TOPOCTR)
        return EclCoord(ecl[0], ecl[1])

    def swe_equator_coord(self, jd) -> EquatorCoord:
        (_, _, _, equator) = swe.nod_aps_ut(jd, self.__swe_code,
                                        swe.NODBIT_FOPOINT | swe.NODBIT_OSCU,
                                        swe.FLG_SWIEPH | swe.FLG_TOPOCTR | swe.FLG_EQUATORIAL)
        return EquatorCoord(equator[0], equator[1])


class FixedCelestial(Celestial):
    """Fixed astronomical objects: stars, galactic and deep space objects"""

    def __init__(self, name: str, swe_code: str):
        self.name = name
        self.__swe_code = swe_code

    def swe_id(self):
        return self.__swe_code

    def is_fixed(self) -> bool:
        return True

    def is_focal_point(self) -> bool:
        return False

    def swe_ecl_coord(self, jd) -> EclCoord:
        (ecl, _, _) = swe.fixstar_ut(self.__swe_code, jd, swe.FLG_SWIEPH | swe.FLG_TOPOCTR)
        return EclCoord(ecl[0], ecl[1])

    def swe_equator_coord(self, jd) -> EquatorCoord:
        (equator, _, _) = swe.fixstar_ut(self.__swe_code, jd, swe.FLG_SWIEPH | swe.FLG_TOPOCTR | swe.FLG_EQUATORIAL)
        return EquatorCoord(equator[0], equator[1])


Planet.Sun = Planet("Sun", swe.SUN)
Planet.Moon = Planet("Moon", swe.MOON)
Planet.Mercury = Planet("Mercury", swe.MERCURY)
Planet.Venus = Planet("Venus", swe.VENUS)
Planet.Earth = Planet("Earth", swe.EARTH)
Planet.Mars = Planet("Mars", swe.MARS)
Planet.Jupiter = Planet("Jupiter", swe.JUPITER)
Planet.Saturn = Planet("Saturn", swe.SATURN)
Planet.Uranus = Planet("Uranus", swe.URANUS)
Planet.Neptune = Planet("Neptune", swe.NEPTUNE)
Planet.Pluto = Planet("Pluto", swe.PLUTO)

Planet.septener = [Planet.Sun, Planet.Mars, Planet.Moon, Planet.Mercury, Planet.Jupiter, Planet.Venus, Planet.Saturn]
Planet.novile = Planet.septener + [Planet.Uranus, Planet.Neptune]

BlackSun.BlackEarth = BlackSun("Black Earth", swe.MOON)
BlackSun.Mercury = BlackSun("BS Mercury", swe.MERCURY)
BlackSun.Venus = BlackSun("BS Venus", swe.VENUS)
BlackSun.Earth = BlackSun("BS Earth", swe.EARTH)
BlackSun.Mars = BlackSun("BS Mars", swe.MARS)
BlackSun.Jupiter = BlackSun("BS Jupiter", swe.JUPITER)
BlackSun.Saturn = BlackSun("BS Saturn", swe.SATURN)
BlackSun.Uranus = BlackSun("BS Uranus", swe.URANUS)
BlackSun.Neptune = BlackSun("BS Neptune", swe.NEPTUNE)



from .primitives import GeoLocation, Angle
from .coords import HorCoord, EclCoord, EquatorCoord
from .zodiac import Zodiac, ZodiacConstell
from .natal import NatalObject, Natal
from .celestials import Celestial, Planet, BlackSun, FixedCelestial

__all__ = [GeoLocation, Angle, HorCoord, EclCoord, EquatorCoord, Zodiac, ZodiacConstell,
           Natal, NatalObject,
           Celestial, Planet, BlackSun, FixedCelestial]


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['astrolog']

package_data = \
{'': ['*']}

install_requires = \
['pyswisseph==2.10.02.0.dev1']

setup_kwargs = {
    'name': 'astrolog',
    'version': '0.1.0',
    'description': 'High-level astrology methods based on Swiss Ephemeris',
    'long_description': None,
    'author': 'Ign. Aiom Ominous',
    'author_email': 'aiom.ominous@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from dms2dec.dms_convert import dms2dec
import math


class Angle:
    """Angle measured in degrees"""

    def __init__(self, *arg, **kwarg):
        if len(arg) > 0:
            deg = arg[0]
            if type(deg) is str:
                self.degrees = dms2dec(deg)
            elif type(deg) is float or type(deg) is int:
                self.degrees = deg
        elif 'radians' in kwarg:
            self.degrees = kwarg['radians'] * 180 / math.pi

    def radians(self):
        return self.degrees * 2 * math.pi / 360

    def aspect(self, orb: float = 0.5):
        if -orb <= self.degrees <= orb:
            return 1
        for div in range(2, 14):
            angle = 360 / div
            if angle - orb <= self.degrees <= angle + orb:
                return div
        return None


class GeoLocation:
    """Location on the earth"""

    def __init__(self, longitude, latitude):
        self.longitude = Angle(longitude)
        self.latitude = Angle(latitude)



from .primitives import Angle


class Zodiac:
    def __init__(self, name: str, symbol: chr, offset: int):
        self.name = name
        self.abbr = name[:3]
        self.symbol = symbol
        self.offset = offset

    @classmethod
    def from_longitude(cls, longitude: Angle):
        for sign in Zodiac.all:
            if sign.offset <= longitude.degrees < sign.offset + 30:
                return sign


class ZodiacConstell:
    def __init__(self, name: str, symbol: chr, from_lng: float, upto_lng: float):
        self.name = name
        self.abbr = name[:3]
        self.symbol = symbol
        self.from_lng = from_lng
        self.upto_lng = upto_lng

    @classmethod
    def from_longitude(cls, longitude: Angle):
        for sign in ZodiacConstell.all:
            if sign.from_lng > sign.upto_lng and (
                0 <= longitude.degrees < sign.upto_lng or
                sign.upto_lng <= longitude.degrees <= 360
            ):
                return sign
            elif sign.from_lng <= longitude.degrees < sign.upto_lng:
                return sign


Zodiac.Aries = Zodiac("Aries", '♈', 0)
Zodiac.Taurus = Zodiac("Taurus", '♉', 30)
Zodiac.Gemini = Zodiac("Gemini", '♊', 60)
Zodiac.Cancer = Zodiac("Cancer", '♋', 90)
Zodiac.Leo = Zodiac("Leo", '♌', 120)
Zodiac.Virgo = Zodiac("Virgo", '♍', 150)
Zodiac.Libra = Zodiac("Libra", '♎', 180)
Zodiac.Scorpio = Zodiac("Scorpio", '♏', 210)
Zodiac.Sagittarius = Zodiac("Sagittarius", '♐', 240)
Zodiac.Capricorn = Zodiac("Capricorn", '♑', 270)
Zodiac.Aquarius = Zodiac("Aquarius", '♒', 300)
Zodiac.Pisces = Zodiac("Pisces", '♓', 330)

Zodiac.all = [Zodiac.Aries, Zodiac.Taurus, Zodiac.Gemini, Zodiac.Cancer, Zodiac.Leo, Zodiac.Virgo,
              Zodiac.Libra, Zodiac.Scorpio, Zodiac.Sagittarius, Zodiac.Capricorn, Zodiac.Aquarius, Zodiac.Pisces]

ZodiacConstell.Aries = ZodiacConstell("Aries", '♈', 29.09, 53.47)
ZodiacConstell.Taurus = ZodiacConstell("Taurus", '♉', 53.47, 90.43)
ZodiacConstell.Gemini = ZodiacConstell("Gemini", '♊', 90.43, 118.26)
ZodiacConstell.Cancer = ZodiacConstell("Cancer", '♋', 118.26, 138.18)
ZodiacConstell.Leo = ZodiacConstell("Leo", '♌', 138.18, 174.16)
ZodiacConstell.Virgo = ZodiacConstell("Virgo", '♍', 174.16, 217.80)
ZodiacConstell.Libra = ZodiacConstell("Libra", '♎', 217.80, 241.14)
ZodiacConstell.Scorpio = ZodiacConstell("Scorpio", '♏', 241.14, 248.04)
ZodiacConstell.Ophiuchus = ZodiacConstell("Ophiuchus", '⛎', 248.04, 266.61)
ZodiacConstell.Sagittarius = ZodiacConstell("Sagittarius", '♐', 266.61, 299.71)
ZodiacConstell.Capricorn = ZodiacConstell("Capricorn", '♑', 299.71, 327.89)
ZodiacConstell.Aquarius = ZodiacConstell("Aquarius", '♒', 327.89, 351.57)
ZodiacConstell.Pisces = ZodiacConstell("Pisces", '♓', 351.57, 29.09)

ZodiacConstell.all = [ZodiacConstell.Aries, ZodiacConstell.Taurus, ZodiacConstell.Gemini, ZodiacConstell.Cancer,
                      ZodiacConstell.Leo, ZodiacConstell.Virgo, ZodiacConstell.Libra, ZodiacConstell.Scorpio,
                      ZodiacConstell.Ophiuchus, ZodiacConstell.Sagittarius, ZodiacConstell.Capricorn,
                      ZodiacConstell.Aquarius, ZodiacConstell.Pisces]



from datetime import datetime, time, timedelta
import swisseph as swe

from .celestials import Celestial
from .primitives import GeoLocation
from .coords import HorCoord, EclCoord
from .zodiac import Zodiac, ZodiacConstell


class NatalObject:
    """Natal object computable type"""

    def __init__(self, obj: Celestial, birth: datetime, place: GeoLocation):
        self.name = obj.name
        self.obj = obj
        self.birth = birth
        self.place = place
        self.__ecl_coord = None
        self.__transits = None

    def julday(self) -> float:
        return swe.julday(self.birth.year, self.birth.month, self.birth.day, self.birth.hour + self.birth.minute / 60.)

    def ecl_coord(self) -> EclCoord:
        if self.__ecl_coord is None:
            self.__ecl_coord = self.obj.swe_ecl_coord(self.julday())
        return self.__ecl_coord

    def hor_coord(self) -> HorCoord:
        coord = self.ecl_coord()
        geopos = (self.place.longitude.degrees, self.place.latitude.degrees, 0.0)
        pos = (coord.longitude.degrees, coord.latitude.degrees, 0.0)
        atpress = 0
        attemp = 0
        (azimuth, true_alt, app_alt) = swe.azalt(self.julday(), swe.ECL2HOR, geopos, atpress, attemp, pos)
        return HorCoord(azimuth, true_alt)

    def sign_pos(self) -> (Zodiac, float):
        return self.ecl_coord().sign_pos()

    def constell_pos(self) -> (ZodiacConstell, float):
        return self.ecl_coord().constell_pos()

    def house_pos(self) -> (int, float):
        return self.hor_coord().house_pos()

    def transits(self) -> dict:
        if self.__transits is None:
            self.__transits = self.obj.transits(self.birth, self.place)
        return self.__transits


class Natal:
    """Natal chart"""

    def __init__(self, person: str, birth: datetime, place: GeoLocation, celestials: [Celestial]):
        self.person = person
        self.birth = birth
        self.place = place
        self.celestials = {obj: NatalObject(obj, birth, place) for obj in celestials}
        swe.set_topo(self.place.longitude.degrees, self.place.latitude.degrees)

    def __iter__(self):
        for cel in self.celestials:
            yield self.celestials[cel]

    def __getitem__(self, celestial: Celestial) -> NatalObject:
        return self.celestials[celestial]

    def aspects(self, orb: float = 1.01, of=None, to=None):
        if type(of) is list:
            list1 = of
        elif isinstance(of, Celestial):
            list1 = [of]
        else:
            list1 = self.celestials

        if type(to) is list:
            list2 = to
        elif isinstance(to, Celestial):
            list2 = [to]
        else:
            list2 = self.celestials

        for (i1, cel1) in enumerate(list1):
            for (i2, cel2) in enumerate(list2):
                if cel1 == cel2:
                    break
                angle = self[cel1].ecl_coord() ^ self[cel2].ecl_coord()
                aspect = angle.aspect(orb=orb)
                if aspect is None:
                    continue
                yield {"first": cel1.name, "second": cel2.name, "aspect": aspect}

    def parans(self, orb: timedelta = timedelta(minutes=5)) -> list:
        events = []
        timeline = {}
        for cel in self:
            transits = cel.transits()
            for transit_name in transits:
                transit_time = transits[transit_name]
                if transit_time is None:
                    continue
                for tm in timeline:
                    if abs(tm - transit_time) <= orb:
                        events.append([
                            timeline[tm],
                            (transit_name, cel.obj)
                        ])
                timeline[transit_time] = (transit_name, cel.obj)
        return events
